#include "Convert.hpp"
#include <algorithm>
#include <limits>

using std::string;
using std::vector;

bool convertMap(const Protocol::Map& protocolMap,
                Map* pMap,
                string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    auto& map = *pMap;

    map.m_numSites = (uint32_t)protocolMap.m_sites.size();
    map.m_numRivers = (uint32_t)protocolMap.m_rivers.size();
    map.m_numMines = (uint32_t)protocolMap.m_mines.size();

    map.m_siteIds.resize(map.m_numSites);
    for (uint32_t iSite = 0; iSite < map.m_numSites; iSite++)
    {
        map.m_siteIds[iSite] = protocolMap.m_sites[iSite].m_id;
    }
    std::sort(map.m_siteIds.begin(), map.m_siteIds.end());
    for (uint32_t iSite = 1; iSite < map.m_numSites; iSite++)
    {
        if (map.m_siteIds[iSite - 1] == map.m_siteIds[iSite])
        {
            *pMsg = "Duplicate site id";
            return false;
        }
    }

    map.m_minCoord.m_x = std::numeric_limits<double>::max();
    map.m_minCoord.m_y = std::numeric_limits<double>::max();
    map.m_maxCoord.m_x = std::numeric_limits<double>::lowest();
    map.m_maxCoord.m_y = std::numeric_limits<double>::lowest();

    map.m_siteCoords.resize(map.m_numSites);
    for (uint32_t iSite = 0; iSite < map.m_numSites; iSite++)
    {
        auto& site = protocolMap.m_sites[iSite];
        uint32_t siteId = site.m_id;
        uint32_t siteIdx = 0;
        if (!map.getSiteIdx(siteId, &siteIdx))
        {
            *pMsg = "Site idx lookup failed";
            return false;
        }
        auto& siteCoord = map.m_siteCoords[siteIdx];
        siteCoord.m_x = site.m_x;
        siteCoord.m_y = site.m_y;

        map.m_minCoord.m_x = std::min(map.m_minCoord.m_x, siteCoord.m_x);
        map.m_minCoord.m_y = std::min(map.m_minCoord.m_y, siteCoord.m_y);
        map.m_maxCoord.m_x = std::max(map.m_maxCoord.m_x, siteCoord.m_x);
        map.m_maxCoord.m_y = std::max(map.m_maxCoord.m_y, siteCoord.m_y);
    }

    map.m_rivers.resize(map.m_numRivers);
    for (uint32_t iRiver = 0; iRiver < map.m_numRivers; iRiver++)
    {
        uint32_t sourceId = protocolMap.m_rivers[iRiver].m_source;
        uint32_t targetId = protocolMap.m_rivers[iRiver].m_target;
        if (sourceId == targetId)
        {
            *pMsg = "River connects site to itself";
            return false;
        }
        uint32_t sourceIdx = 0;
        uint32_t targetIdx = 0;
        if (!map.getSiteIdx(sourceId, &sourceIdx) ||
            !map.getSiteIdx(targetId, &targetIdx))
        {
            *pMsg = "Site idx lookup failed";
            return false;
        }
        map.m_rivers[iRiver] = River(sourceIdx, targetIdx);
    }
    std::sort(map.m_rivers.begin(), map.m_rivers.end());
    for (uint32_t iRiver = 1; iRiver < map.m_numRivers; iRiver++)
    {
        if (map.m_rivers[iRiver - 1] == map.m_rivers[iRiver])
        {
            *pMsg = "Duplicate river";
            return false;
        }
    }

    map.initSiteRivers();

    map.m_mines.resize(map.m_numMines);
    for (uint32_t iMine = 0; iMine < map.m_numMines; iMine++)
    {
        uint32_t siteId = protocolMap.m_mines[iMine];
        uint32_t siteIdx = 0;
        if (!map.getSiteIdx(siteId, &siteIdx))
        {
            *pMsg = "Site idx lookup failed";
            return false;
        }
        map.m_mines[iMine] = siteIdx;
    }
    std::sort(map.m_mines.begin(), map.m_mines.end());
    for (uint32_t iMine = 1; iMine < map.m_numMines; iMine++)
    {
        if (map.m_mines[iMine - 1] == map.m_mines[iMine])
        {
            *pMsg = "Duplicate mine";
            return false;
        }
    }

    map.initIsMine();

    return true;
}
