#ifndef MAP_HPP
#define MAP_HPP

#include "Common.hpp"

class Coord
{
public:
    Coord() : m_x(0.0), m_y(0.0) {}
    Coord(double x, double y) : m_x(x), m_y(y) {}

public:
    double m_x;
    double m_y;
};

class River
{
public:
    River() : m_source(0), m_target(0) {}
    River(uint32_t source, uint32_t target) :
        m_source(std::min(source, target)),
        m_target(std::max(source, target))
    {
    }

    bool operator==(const River& other) const
    {
        return m_source == other.m_source && m_target == other.m_target;
    }
    bool operator!=(const River& other) const
    {
        return m_source != other.m_source || m_target != other.m_target;
    }
    bool operator<(const River& other) const
    {
        if (m_source != other.m_source) return m_source < other.m_source;
        return m_target < other.m_target;
    }

public:
    uint32_t m_source;
    uint32_t m_target;
};

class SiteRiver
{
public:
    SiteRiver() : m_source(0), m_target(0), m_river(0) {}
    SiteRiver(uint32_t source, uint32_t target, uint32_t river) :
        m_source(source), m_target(target), m_river(river) {}

    bool operator==(const SiteRiver& other) const
    {
        return m_source == other.m_source &&
            m_target == other.m_target &&
            m_river == other.m_river;
    }
    bool operator!=(const SiteRiver& other) const
    {
        return m_source != other.m_source ||
            m_target != other.m_target ||
            m_river != other.m_river;
    }
    bool operator<(const SiteRiver& other) const
    {
        if (m_source != other.m_source) return m_source < other.m_source;
        if (m_target != other.m_target) return m_target < other.m_target;
        return m_river < other.m_river;
    }

public:
    uint32_t m_source;
    uint32_t m_target;
    uint32_t m_river;
};

class SiteRiverBounds
{
public:
    SiteRiverBounds() : m_low(0), m_high(0) {}
    SiteRiverBounds(uint32_t low, uint32_t high) :
        m_low(low), m_high(high) {}

public:
    uint32_t m_low;
    uint32_t m_high;
};

class Map
{
public:
    Map();
    ~Map();

    uint32_t getSiteId(uint32_t siteIdx) const;
    bool getSiteIdx(uint32_t siteId, uint32_t* pSiteIdx = nullptr) const;

    bool getRiverIdx(const River& river, uint32_t* pRiverIdx = nullptr) const;

    void initSiteRivers();
    void initIsMine();

    void serialize(std::vector<uint32_t>* pData) const;
    bool deserialize(const std::vector<uint32_t>& data, size_t& pos);

public:
    uint32_t m_numSites;
    uint32_t m_numRivers;
    uint32_t m_numMines;
    std::vector<River> m_rivers;
    std::vector<SiteRiver> m_siteRivers;
    std::vector<SiteRiverBounds> m_siteRiverBounds;
    std::vector<uint32_t> m_mines;
    std::vector<uint8_t> m_isMine;

    // Not used for solving
    std::vector<uint32_t> m_siteIds; // sorted
    std::vector<Coord> m_siteCoords;
    Coord m_minCoord;
    Coord m_maxCoord;
};

#endif
