#ifndef PROGRAM_HPP
#define PROGRAM_HPP

#include "Common.hpp"
#include "Game.hpp"
#include "Protocol.hpp"

class Solver;

class Program
{
public:
    Program(Solver& solver,
            const std::string& programName,
            const std::string& punterName = std::string());
    ~Program();

    int run(int argc, char *argv[]);

private:
    void usage();
    int runOnline(const std::string& host,
                  const std::string& port);
    int runOffline();

    bool processSetup(const Protocol::Setup& setup,
                      Protocol::Ready* pReady);
    bool processPlay(const Protocol::Play& play,
                     Protocol::Action* pAction);
    bool processStop(const Protocol::Stop& stop);
    bool applyMoves(const std::vector<Protocol::Move>& moves);
    void serialize(std::string* pState) const;
    bool deserialize(const std::string& state);

private:
    std::string m_programName;
    std::string m_punterName;
    Solver& m_solver;
    Game m_game;
    bool m_online;
};

#endif
