#include "Protocol.hpp"
#include "Cleanup.hpp"
#include <jansson.h>

// sudo apt-get install libjansson-dev

using std::string;
using std::vector;

namespace Protocol
{
    bool formatHandshake(const ClientHandshake& handshake,
                         string* pStrHandshake,
                         string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        json_t* jHandshake = json_object();
        if (!jHandshake)
        {
            *pMsg = "Create handshake failed";
            return false;
        }
        Cleanup cleanupHandshake([&](){ json_decref(jHandshake); });

        if (json_object_set_new(jHandshake, "me",
                                json_string(handshake.m_me.c_str())) != 0)
        {
            *pMsg = "Set name failed";
            return false;
        }

        char* rawHandshake = json_dumps(jHandshake, 0);
        if (!rawHandshake)
        {
            *pMsg = "Dump handshake failed";
            return false;
        }
        Cleanup cleanupRawHandshake([&](){ free(rawHandshake); });

        *pStrHandshake = rawHandshake;
        return true;
    }

    bool parseHandshake(const string& strHandshake,
                        ServerHandshake* pHandshake,
                        string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        json_error_t jerr;
        json_t* jHandshake = json_loads(strHandshake.c_str(), 0, &jerr);
        if (!jHandshake)
        {
            *pMsg = "Load handshake failed";
            return false;
        }
        Cleanup cleanupHandshake([&](){ json_decref(jHandshake); });
        if (!json_is_object(jHandshake))
        {
            *pMsg = "Handshake not object";
            return false;
        }

        auto& handshake = *pHandshake;

        {
            json_t* jYou = json_object_get(jHandshake, "you");
            if (!jYou)
            {
                *pMsg = "Error getting handshake name";
                return false;
            }
            if (!json_is_string(jYou))
            {
                *pMsg = "Handshake name not string";
                return false;
            }
            const char* rawYou = json_string_value(jYou);
            handshake.m_you = rawYou;
        }

        return true;
    }

    bool parseMap(json_t* jMap,
                  Map* pMap,
                  string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_object(jMap))
        {
            *pMsg = "Map not object";
            return false;
        }

        auto& map = *pMap;

        {
            json_t* jSites = json_object_get(jMap, "sites");
            if (!jSites)
            {
                *pMsg = "Error getting sites";
                return false;
            }
            if (!json_is_array(jSites))
            {
                *pMsg = "Sites not array";
                return false;
            }

            map.m_sites.clear();

            uint32_t siteCount = (uint32_t)json_array_size(jSites);
            for (uint32_t iSite = 0; iSite < siteCount; iSite++)
            {
                json_t* jSite = json_array_get(jSites, iSite);
                if (!jSite)
                {
                    *pMsg = "Error getting site";
                    return false;
                }
                if (!json_is_object(jSite))
                {
                    *pMsg = "Site not object";
                    return false;
                }

                map.m_sites.emplace_back();
                auto& site = map.m_sites.back();

                {
                    json_t* jId = json_object_get(jSite, "id");
                    if (!jId)
                    {
                        *pMsg = "Error getting site id";
                        return false;
                    }
                    if (!json_is_integer(jId))
                    {
                        *pMsg = "Site id not integer";
                        return false;
                    }
                    site.m_id = (uint32_t)json_integer_value(jId);
                }

                {
                    json_t* jX = json_object_get(jSite, "x");
                    json_t* jY = json_object_get(jSite, "y");
                    if (!jX && !jY)
                    {
                    }
                    else
                    {
                        if (!jX)
                        {
                            *pMsg = "Error getting site x";
                            return false;
                        }
                        else
                        {
                            if (!json_is_number(jX))
                            {
                                *pMsg = "Site x not number";
                                return false;
                            }
                            site.m_x = json_number_value(jX);
                        }

                        if (!jY)
                        {
                            *pMsg = "Error getting site y";
                            return false;
                        }
                        else
                        {
                            if (!json_is_number(jY))
                            {
                                *pMsg = "Site y not number";
                                return false;
                            }
                            site.m_y = json_number_value(jY);
                        }
                    }
                }
            }
        }

        {
            json_t* jRivers = json_object_get(jMap, "rivers");
            if (!jRivers)
            {
                *pMsg = "Error getting rivers";
                return false;
            }
            if (!json_is_array(jRivers))
            {
                *pMsg = "Rivers not array";
                return false;
            }

            map.m_rivers.clear();

            uint32_t riverCount = (uint32_t)json_array_size(jRivers);
            for (uint32_t iRiver = 0; iRiver < riverCount; iRiver++)
            {
                json_t* jRiver = json_array_get(jRivers, iRiver);
                if (!jRiver)
                {
                    *pMsg = "Error getting river";
                    return false;
                }
                if (!json_is_object(jRiver))
                {
                    *pMsg = "River not object";
                    return false;
                }

                map.m_rivers.emplace_back();
                auto& river = map.m_rivers.back();

                {
                    json_t* jSource = json_object_get(jRiver, "source");
                    if (!jSource)
                    {
                        *pMsg = "Error getting river source";
                        return false;
                    }
                    if (!json_is_integer(jSource))
                    {
                        *pMsg = "River source not integer";
                        return false;
                    }
                    river.m_source = (uint32_t)json_integer_value(jSource);
                }

                {
                    json_t* jTarget = json_object_get(jRiver, "target");
                    if (!jTarget)
                    {
                        *pMsg = "Error getting river target";
                        return false;
                    }
                    if (!json_is_integer(jTarget))
                    {
                        *pMsg = "River target not integer";
                        return false;
                    }
                    river.m_target = (uint32_t)json_integer_value(jTarget);
                }
            }
        }

        {
            json_t* jMines = json_object_get(jMap, "mines");
            if (!jMines)
            {
                *pMsg = "Error getting mines";
                return false;
            }
            if (!json_is_array(jMines))
            {
                *pMsg = "Mines not array";
                return false;
            }

            map.m_mines.clear();

            uint32_t mineCount = (uint32_t)json_array_size(jMines);
            for (uint32_t iMine = 0; iMine < mineCount; iMine++)
            {
                json_t* jMine = json_array_get(jMines, iMine);
                if (!jMine)
                {
                    *pMsg = "Error getting mine";
                    return false;
                }
                if (!json_is_integer(jMine))
                {
                    *pMsg = "Mine not integer";
                    return false;
                }

                map.m_mines.emplace_back();
                auto& mine = map.m_mines.back();

                mine = (uint32_t)json_integer_value(jMine);
            }
        }

        return true;
    }

    bool parseSetup(json_t* jSetup,
                    Setup* pSetup,
                    string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_object(jSetup))
        {
            *pMsg = "Setup not object";
            return false;
        }

        auto& setup = *pSetup;

        {
            json_t* jPunter = json_object_get(jSetup, "punter");
            if (!jPunter)
            {
                *pMsg = "Error getting setup punter";
                return false;
            }
            if (!json_is_integer(jPunter))
            {
                *pMsg = "Setup punter not integer";
                return false;
            }
            setup.m_punter = (uint32_t)json_integer_value(jPunter);
        }

        {
            json_t* jPunters = json_object_get(jSetup, "punters");
            if (!jPunters)
            {
                *pMsg = "Error getting setup punters";
                return false;
            }
            if (!json_is_integer(jPunters))
            {
                *pMsg = "Setup punters not integer";
                return false;
            }
            setup.m_punters = (uint32_t)json_integer_value(jPunters);
        }

        {
            json_t* jMap = json_object_get(jSetup, "map");
            if (!jMap)
            {
                *pMsg = "Error getting setup map";
                return false;
            }

            if (!parseMap(jMap, &setup.m_map, pMsg))
            {
                return false;
            }
        }

        {
            json_t* jSettings = json_object_get(jSetup, "settings");
            if (jSettings)
            {
                if (!json_is_object(jSettings))
                {
                    *pMsg = "Settings not object";
                    return false;
                }

                {
                    json_t* jFutures = json_object_get(jSettings, "futures");
                    if (!jFutures)
                    {
                        *pMsg = "Error getting settings futures";
                        return false;
                    }
                    if (!json_is_boolean(jFutures))
                    {
                        *pMsg = "Settings futures not boolean";
                        return false;
                    }

                    setup.m_useFutures = !!json_is_true(jFutures);
                }

            }
        }

        return true;
    }

    bool parseMoves(json_t* jMoves,
                    vector<Move>* pMoves,
                    std::string* pMsg = nullptr)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_array(jMoves))
        {
            *pMsg = "Moves not array";
            return false;
        }

        auto& moves = *pMoves;

        uint32_t moveCount = (uint32_t)json_array_size(jMoves);
        for (uint32_t iMove = 0; iMove < moveCount; iMove++)
        {
            json_t* jMove = json_array_get(jMoves, iMove);
            if (!jMove)
            {
                *pMsg = "Error getting move";
                return false;
            }
            if (!json_is_object(jMove))
            {
                *pMsg = "Move not object";
                return false;
            }

            moves.emplace_back();
            auto& move = moves.back();

            json_t* jPass = json_object_get(jMove, "pass");
            json_t* jClaim = json_object_get(jMove, "claim");
            if (!jPass && !jClaim)
            {
                *pMsg = "Pass and claim both not found";
                return false;
            }
            if (jPass && jClaim)
            {
                *pMsg = "Pass and claim both found";
                return false;
            }

            if (jPass)
            {
                move.m_moveType = MoveType::Pass;

                {
                    json_t* jPunter = json_object_get(jPass, "punter");
                    if (!jPunter)
                    {
                        *pMsg = "Error getting move punter";
                        return false;
                    }
                    if (!json_is_integer(jPunter))
                    {
                        *pMsg = "Move punter not integer";
                        return false;
                    }
                    move.m_punter = (uint32_t)json_integer_value(jPunter);
                }
            }
            else
            {
                move.m_moveType = MoveType::Claim;

                {
                    json_t* jPunter = json_object_get(jClaim, "punter");
                    if (!jPunter)
                    {
                        *pMsg = "Error getting move punter";
                        return false;
                    }
                    if (!json_is_integer(jPunter))
                    {
                        *pMsg = "Move punter not integer";
                        return false;
                    }
                    move.m_punter = (uint32_t)json_integer_value(jPunter);
                }

                {
                    json_t* jSource = json_object_get(jClaim, "source");
                    if (!jSource)
                    {
                        *pMsg = "Error getting move source";
                        return false;
                    }
                    if (!json_is_integer(jSource))
                    {
                        *pMsg = "Move source not integer";
                        return false;
                    }
                    move.m_source = (uint32_t)json_integer_value(jSource);
                }

                {
                    json_t* jTarget = json_object_get(jClaim, "target");
                    if (!jTarget)
                    {
                        *pMsg = "Error getting move target";
                        return false;
                    }
                    if (!json_is_integer(jTarget))
                    {
                        *pMsg = "Move target not integer";
                        return false;
                    }
                    move.m_target = (uint32_t)json_integer_value(jTarget);
                }
            }
        }

        return true;
    }

    bool parsePlay(bool online,
                   json_t* jPlay,
                   Play* pPlay,
                   string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_object(jPlay))
        {
            *pMsg = "Play not object";
            return false;
        }

        auto& play = *pPlay;

        {
            json_t* jMove = json_object_get(jPlay, "move");
            if (!jMove)
            {
                *pMsg = "Error getting play move";
                return false;
            }
            if (!json_is_object(jMove))
            {
                *pMsg = "Play move not object";
                return false;
            }

            {
                json_t* jMoves = json_object_get(jMove, "moves");
                if (!jMoves)
                {
                    *pMsg = "Error getting moves";
                    return false;
                }

                if (!parseMoves(jMoves, &play.m_moves, pMsg))
                {
                    return false;
                }

            }
        }

        if (!online)
        {
            json_t* jState = json_object_get(jPlay, "state");
            if (!jState)
            {
                *pMsg = "Error getting state";
                return false;
            }
            if (!json_is_string(jState))
            {
                *pMsg = "Play state not object";
                return false;
            }
            const char* rawState = json_string_value(jState);
            play.m_state = rawState;
        }

        return true;
    }

    bool parseStop(bool online,
                   json_t* jStopRequest,
                   Stop* pStop,
                   string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_object(jStopRequest))
        {
            *pMsg = "Stop not object";
            return false;
        }

        auto& stop = *pStop;

        {
            json_t* jStop = json_object_get(jStopRequest, "stop");
            if (!jStop)
            {
                *pMsg = "Error getting stop stop";
                return false;
            }
            if (!json_is_object(jStop))
            {
                *pMsg = "Stop stop not object";
                return false;
            }

            {
                json_t* jMoves = json_object_get(jStop, "moves");
                if (!jMoves)
                {
                    *pMsg = "Error getting moves";
                    return false;
                }

                if (!parseMoves(jMoves, &stop.m_moves, pMsg))
                {
                    return false;
                }

            }

            {
                json_t* jScores = json_object_get(jStop, "scores");
                if (!jScores)
                {
                    *pMsg = "Error getting scores";
                    return false;
                }
                if (!json_is_array(jScores))
                {
                    *pMsg = "Scores not array";
                    return false;
                }

                stop.m_scores.clear();

                uint32_t scoreCount = (uint32_t)json_array_size(jScores);
                for (uint32_t iScore = 0; iScore < scoreCount; iScore++)
                {
                    json_t* jScore = json_array_get(jScores, iScore);
                    if (!jScore)
                    {
                        *pMsg = "Error getting score";
                        return false;
                    }
                    if (!json_is_object(jScore))
                    {
                        *pMsg = "Score not object";
                        return false;
                    }

                    stop.m_scores.emplace_back();
                    auto& score = stop.m_scores.back();

                    {
                        json_t* jPunter = json_object_get(jScore, "punter");
                        if (!jPunter)
                        {
                            *pMsg = "Error getting score punter";
                            return false;
                        }
                        if (!json_is_integer(jPunter))
                        {
                            *pMsg = "Score punter not integer";
                            return false;
                        }
                        score.m_punter = (uint32_t)json_integer_value(jPunter);
                    }

                    {
                        json_t* jValue = json_object_get(jScore, "score");
                        if (!jValue)
                        {
                            *pMsg = "Error getting score value";
                            return false;
                        }
                        if (!json_is_integer(jValue))
                        {
                            *pMsg = "Score value not integer";
                            return false;
                        }
                        score.m_score = (int64_t)json_integer_value(jValue);
                    }
                }
            }
        }

        if (!online)
        {
            json_t* jState = json_object_get(jStopRequest, "state");
            if (!jState)
            {
                *pMsg = "Error getting state";
                return false;
            }
            if (!json_is_string(jState))
            {
                *pMsg = "Stop state not object";
                return false;
            }
            const char* rawState = json_string_value(jState);
            stop.m_state = rawState;
        }

        return true;
    }

    bool parseTimeout(json_t* jTimeout,
                      Timeout* pTimeout,
                      string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        if (!json_is_object(jTimeout))
        {
            *pMsg = "Timeout not object";
            return false;
        }

        auto& timeout = *pTimeout;

        {
            json_t* jValue = json_object_get(jTimeout, "timeout");
            if (!jValue)
            {
                *pMsg = "Error getting timeout value";
                return false;
            }
            if (!json_is_number(jValue))
            {
                *pMsg = "Timeout value not number";
                return false;
            }
            timeout.m_timeout = json_number_value(jValue);
        }

        return true;
    }

    bool parseMessage(bool online,
                      const string& strMessage,
                      MessageType* pMessageType,
                      Setup* pSetup,
                      Play* pPlay,
                      Stop* pStop,
                      Timeout* pTimeout,
                      string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        MessageType messageType;
        if (!pMessageType)
        {
            pMessageType = &messageType;
        }

        json_error_t jerr;
        json_t* jMessage = json_loads(strMessage.c_str(), 0, &jerr);
        if (!jMessage)
        {
            *pMsg = "Load message failed";
            return false;
        }
        Cleanup cleanupMessage([&](){ json_decref(jMessage); });
        if (!json_is_object(jMessage))
        {
            *pMsg = "Setup not object";
            return false;
        }

        if (json_object_get(jMessage, "punter"))
        {
            std::unique_ptr<Setup> uSetup;
            if (!pSetup)
            {
                uSetup.reset(new Setup);
                pSetup = uSetup.get();
            }

            if (!parseSetup(jMessage, pSetup, pMsg))
            {
                return false;
            }

            *pMessageType = MessageType::Setup;
            return true;
        }

        if (json_object_get(jMessage, "move"))
        {
            std::unique_ptr<Play> uPlay;
            if (!pPlay)
            {
                uPlay.reset(new Play);
                pPlay = uPlay.get();
            }

            if (!parsePlay(online, jMessage, pPlay, pMsg))
            {
                return false;
            }

            *pMessageType = MessageType::Play;
            return true;
        }

        if (json_object_get(jMessage, "stop"))
        {
            std::unique_ptr<Stop> uStop;
            if (!pStop)
            {
                uStop.reset(new Stop);
                pStop = uStop.get();
            }

            if (!parseStop(online, jMessage, pStop, pMsg))
            {
                return false;
            }

            *pMessageType = MessageType::Stop;
            return true;
        }

        if (json_object_get(jMessage, "timeout"))
        {
            std::unique_ptr<Timeout> uTimeout;
            if (!pTimeout)
            {
                uTimeout.reset(new Timeout);
                pTimeout = uTimeout.get();
            }

            if (!parseTimeout(jMessage, pTimeout, pMsg))
            {
                return false;
            }

            *pMessageType = MessageType::Timeout;
            return true;
        }

        *pMsg = "Unknown message type";
        return false;
    }

   
    bool formatReady(bool online,
                     bool useFutures,
                     const Ready& ready,
                     string* pStrReady,
                     string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        json_t* jReady = json_object();
        if (!jReady)
        {
            *pMsg = "Create ready failed";
            return false;
        }
        Cleanup cleanupReady([&](){ json_decref(jReady); });

        if (json_object_set_new(jReady, "ready",
                                json_integer(ready.m_ready)) != 0)
        {
            *pMsg = "Set ready failed";
            return false;
        }

        if (useFutures)
        {
            json_t* jFutures = json_array();
            if (!jFutures)
            {
                *pMsg = "Create futures failed";
                return false;
            }
            Cleanup cleanupFutures([&](){ json_decref(jFutures); });

            for (auto& future : ready.m_futures)
            {
                json_t* jFuture = json_object();
                if (!jFuture)
                {
                    *pMsg = "Create future failed";
                    return false;
                }
                Cleanup cleanupFuture([&](){ json_decref(jFuture); });

                if (json_object_set_new(jFuture, "source",
                                        json_integer(future.m_source)) != 0)
                {
                    *pMsg = "Set source failed";
                    return false;
                }

                if (json_object_set_new(jFuture, "target",
                                        json_integer(future.m_target)) != 0)
                {
                    *pMsg = "Set target failed";
                    return false;
                }

                if (json_array_append(jFutures,
                                      jFuture) != 0)
                {
                    *pMsg = "Append future failed";
                    return false;
                }
            }

            if (json_object_set(jReady, "futures",
                                jFutures) != 0)
            {
                *pMsg = "Set futures failed";
                return false;
            }
        }

        if (!online)
        {
            if (json_object_set_new(jReady, "state",
                                    json_string(ready.m_state.c_str())) != 0)
            {
                *pMsg = "Set state failed";
                return false;
            }
        }

        char* rawReady = json_dumps(jReady, 0);
        if (!rawReady)
        {
            *pMsg = "Dump ready failed";
            return false;
        }
        Cleanup cleanupRawReady([&](){ free(rawReady); });

        *pStrReady = rawReady;
        return true;
    }

    bool formatAction(bool online,
                      const Action& action,
                      string* pStrAction,
                      string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        json_t* jAction = json_object();
        if (!jAction)
        {
            *pMsg = "Create action failed";
            return false;
        }
        Cleanup cleanupAction([&](){ json_decref(jAction); });

        auto& move = action.m_move;

        if (move.m_moveType == MoveType::Pass)
        {
            json_t* jPass = json_object();
            if (!jPass)
            {
                *pMsg = "Create pass failed";
                return false;
            }
            Cleanup cleanupPass([&](){ json_decref(jPass); });

            if (json_object_set_new(jPass, "punter",
                                    json_integer(move.m_punter)) != 0)
            {
                *pMsg = "Set punter failed";
                return false;
            }

            if (json_object_set(jAction, "pass",
                                jPass) != 0)
            {
                *pMsg = "Set pass failed";
                return false;
            }
        }
        else
        {
            json_t* jClaim = json_object();
            if (!jClaim)
            {
                *pMsg = "Create claim failed";
                return false;
            }
            Cleanup cleanupClaim([&](){ json_decref(jClaim); });

            if (json_object_set_new(jClaim, "punter",
                                    json_integer(move.m_punter)) != 0)
            {
                *pMsg = "Set punter failed";
                return false;
            }

            if (json_object_set_new(jClaim, "source",
                                    json_integer(move.m_source)) != 0)
            {
                *pMsg = "Set source failed";
                return false;
            }

            if (json_object_set_new(jClaim, "target",
                                    json_integer(move.m_target)) != 0)
            {
                *pMsg = "Set target failed";
                return false;
            }

            if (json_object_set(jAction, "claim",
                                jClaim) != 0)
            {
                *pMsg = "Set claim failed";
                return false;
            }
        }

        if (!online)
        {
            if (json_object_set_new(jAction, "state",
                                    json_string(action.m_state.c_str())) != 0)
            {
                *pMsg = "Set state failed";
                return false;
            }
        }

        char* rawAction = json_dumps(jAction, 0);
        if (!rawAction)
        {
            *pMsg = "Dump action failed";
            return false;
        }
        Cleanup cleanupRawAction([&](){ free(rawAction); });

        *pStrAction = rawAction;
        return true;
    }

    bool parseMap(const string& strMap,
                  Map* pMap,
                  string* pMsg)
    {
        string msg;
        if (!pMsg)
        {
            pMsg = &msg;
        }

        json_error_t jerr;
        json_t* jMap = json_loads(strMap.c_str(), 0, &jerr);
        if (!jMap)
        {
            *pMsg = "Load map failed";
            return false;
        }
        Cleanup cleanupMap([&](){ json_decref(jMap); });

        if (!parseMap(jMap, pMap, pMsg))
        {
            return false;
        }

        return true;
    }
}
