#ifndef PROTOCOL_HPP
#define PROTOCOL_HPP

#include "Common.hpp"

class Game;

namespace Protocol
{
    class ClientHandshake
    {
    public:
        std::string m_me;
    };

    class ServerHandshake
    {
    public:
        std::string m_you;
    };

    class Site
    {
    public:
        uint32_t m_id = 0;
        double m_x = 0.0;
        double m_y = 0.0;
    };

    class River
    {
    public:
        uint32_t m_source = 0;
        uint32_t m_target = 0;
    };

    class Map
    {
    public:
        std::vector<Site> m_sites;
        std::vector<River> m_rivers;
        std::vector<uint32_t> m_mines;
    };

    class Setup
    {
    public:
        uint32_t m_punter = 0;
        uint32_t m_punters = 0;
        Map m_map;
        bool m_useFutures = false;
    };

    class Future
    {
    public:
        uint32_t m_source = 0;
        uint32_t m_target = 0;
    };

    class Ready
    {
    public:
        uint32_t m_ready = 0;
        std::vector<Future> m_futures;
        std::string m_state;
    };

    enum class MoveType : uint32_t
    {
        Pass,
        Claim
    };

    class Move
    {
    public:
        MoveType m_moveType = MoveType::Pass;
        uint32_t m_punter = 0;
        uint32_t m_source = 0;
        uint32_t m_target = 0;
    };

    class Play
    {
    public:
        std::vector<Move> m_moves;
        std::string m_state;
    };

    class Action
    {
    public:
        Move m_move;
        std::string m_state;
    };

    class Score
    {
    public:
        uint32_t m_punter = 0;
        int64_t m_score = 0;
    };

    class Stop
    {
    public:
        std::vector<Move> m_moves;
        std::vector<Score> m_scores;
        std::string m_state;
    };

    class Timeout
    {
    public:
        double m_timeout;
    };

    bool formatHandshake(const ClientHandshake& handshake,
                         std::string* pStrHandshake,
                         std::string* pMsg = nullptr);
    bool parseHandshake(const std::string& strHandshake,
                        ServerHandshake* pHandshake,
                        std::string* pMsg = nullptr);

    enum class MessageType : uint32_t
    {
        Setup,
        Play,
        Stop,
        Timeout
    };

    bool parseMessage(bool online,
                      const std::string& strMessage,
                      MessageType* pMessageType,
                      Setup* pSetup,
                      Play* pPlay,
                      Stop* pStop,
                      Timeout* pTimeout,
                      std::string* pMsg = nullptr);
                      
    bool formatReady(bool online,
                     bool useFutures,
                     const Ready& ready,
                     std::string* pStrReady,
                     std::string* pMsg = nullptr);

    bool formatAction(bool online,
                      const Action& action,
                      std::string* pStrAction,
                      std::string* pMsg = nullptr);

    bool parseMap(const std::string& strMap,
                  Map* pMap,
                  std::string* pMsg = nullptr);
}

#endif
